// ignore_for_file: omit_local_variable_types

part of pdfio;

class PPages extends PObject {
  PDocument document;
  PPages(this.document) : super(document, '/Pages');
  @override
  void _prepare() {
    super._prepare();
    params['/Kids'] = PArray.fromObjects(document.pages);
    params['/Count'] = PNum(document.pages.length);
  }
}

class PRoot extends PObject {
  PDocument document;
  PRoot(this.document) : super(document, '/Catalog');
  Outline outlines;
  @override
  void _prepare() {
    super._prepare();
    params['/Version'] = PName('/${PDocument.version}');
    params['/Pages'] = document.page.idref();
    if (outlines != null && outlines.outlines.isNotEmpty) {
      params['/Outlines'] = outlines.idref();
    }
  }
}

class PPage extends PObject {
  PPage(PDocument document, {this.pageFormat = PPageFormat.standard}) : super(document, '/Page') {
    document.pages.add(this);
  }

  PPageFormat pageFormat;

  List<POStream> contents = <POStream>[];

  PObject thumbnail;

  bool isolatedTransparency = false;

  bool knockoutTransparency = false;

  List<Annotation> annotations = <Annotation>[];

  final Map<String, PFont> fonts = <String, PFont>{};
  final Map<String, PdfShading> shading = <String, PdfShading>{};

  final Map<String, XObject> xObjects = <String, XObject>{};

  PdfGraphics getGraphics() {
    final POStream stream = POStream(document);
    final PdfGraphics g = PdfGraphics(this, stream.stream);
    contents.add(stream);
    return g;
  }

  void addAnnotation(PObject ob) {
    annotations.add(ob);
  }

  /// @param os OutputStream to send the object to
  @override
  void _prepare() {
    super._prepare();

    params['/Parent'] = document.page.idref();
    params['/MediaBox'] = PArray.fromNum(<double>[0, 0, pageFormat.width, pageFormat.height]);

    // the /Contents pages object
    if (contents.isNotEmpty) {
      if (contents.length == 1) {
        params['/Contents'] = contents.first.idref();
      } else {
        params['/Contents'] = PArray.fromObjects(contents);
      }
    }

    final PDict resources = PDict();

    resources['/ProcSet'] = PArray([PName('/PDF'), PName('/Text'), PName('/ImageB'), PName('/ImageC')]);

    if (fonts.isNotEmpty) resources['/Font'] = PDict.fromObjectMap(fonts);
    if (shading.isNotEmpty) resources['/Shading'] = PDict.fromObjectMap(shading);
    if (xObjects.isNotEmpty) resources['/XObject'] = PDict.fromObjectMap(xObjects);

    if (document.hasGraphic) {
      // Declare Transparency Group settings
      params['/Group'] = PDict(<String, PType>{
        '/Type': const PName('/Group'),
        '/S': const PName('/Transparency'),
        '/CS': const PName('/DeviceRGB'),
        '/I': PBool(isolatedTransparency),
        '/K': PBool(knockoutTransparency),
      });
      resources['/ExtGState'] = document.graphicStates.idref();
    }

    params['/Resources'] = resources;

    if (thumbnail != null) params['/Thumb'] = thumbnail.idref();
    if (annotations.isNotEmpty) params['/Annots'] = PArray.fromObjects(annotations);
  }

  void fillImage(File imFile, {PPageFormat format = PPageFormat.a4}) {
    var image = PImage.fromImage(document, image: decodeImage(imFile.readAsBytesSync()));
    var canvas = getGraphics();
    var destinationRect = PRect(0.0, 0.0, format.width, format.height);
    var sourceRect = PRect(0.0, 0.0, image.width.toDouble(), image.height.toDouble());
    var fw = destinationRect.width / sourceRect.width;
    var fh = destinationRect.height / sourceRect.height;
    canvas.saveContext();
    canvas.drawRect(destinationRect.x, destinationRect.y, destinationRect.width, destinationRect.height);
    canvas.clipPath();
    canvas.drawImage(image, destinationRect.x - sourceRect.x * fw, destinationRect.y - sourceRect.y * fh, image.width.toDouble() * fw, image.height.toDouble() * fh);
    canvas.restoreContext();
  }
}

class PPageFormat {

  static const double point = 1.0;
  static const double inch = 72.0;
  static const double cm = inch / 2.54;
  static const double mm = inch / 25.4;

  const PPageFormat(this.width, this.height,
      {double marginTop = 0.0,
        double marginBottom = 0.0,
        double marginLeft = 0.0,
        double marginRight = 0.0,
        double margin})
      : assert(width > 0),
        assert(height > 0),
        top = margin ?? marginTop,
        bottom = margin ?? marginBottom,
        left = margin ?? marginLeft,
        right = margin ?? marginRight;

  static const PPageFormat a3 = PPageFormat(29.7 * cm, 42 * cm, margin: 2.0 * cm);
  static const PPageFormat a4 = PPageFormat(21.0 * cm, 29.7 * cm, margin: 2.0 * cm);
  static const PPageFormat a5 = PPageFormat(14.8 * cm, 21.0 * cm, margin: 2.0 * cm);

  static const PPageFormat standard = a4;

  final double width;
  final double height;

  final double top;
  final double bottom;
  final double left;
  final double right;

  @override
  String toString() {
    return 'Page ${width}x$height margins:$left, $top, $right, $bottom';
  }
}

