
part of pdfio;

/// pdf data type
abstract class PType {
  const PType();
  void output(PStream s);
  PStream _toStream() {
    var s = PStream();
    output(s);
    return s;
  }
  @override
  String toString() {
    return String.fromCharCodes(_toStream().output());
  }
  @visibleForTesting
  Uint8List toList() {
    return _toStream().output();
  }
}

class PBool extends PType {
  const PBool(this.value);
  final bool value;
  @override
  void output(PStream s) {
    s.putString(value ? 'true' : 'false');
  }
}

class PNum extends PType {
  const PNum(this.value)
      : assert(value != null),
        assert(value != double.infinity),
        assert(value != double.nan),
        assert(value != double.negativeInfinity);

  static const int precision = 5;

  final num value;

  @override
  void output(PStream s) {
    if (value is int) {
      s.putString(value.toInt().toString());
    } else {
      s.putString(value.toStringAsFixed(precision));
    }
  }
}

class PNums extends PType {
  PNums(this.values) : assert(values != null);

  final List<num> values;

  @override
  void output(PStream s) {
    for (var n = 0; n < values.length; n++) {
      if (n > 0) {
        s.putByte(0x20);
      }
      PNum(values[n]).output(s);
    }
  }
}


class PString extends PType {
  const PString(this.value, [this.format = false]);

  factory PString.fromString(String value) {
    return PString(_string(value));
  }

  factory PString.fromDate(DateTime date) {
    return PString(_date(date));
  }

  final Uint8List value;

  final bool format;

  static Uint8List _string(String value) {
    try {
      return latin1.encode(value);
    } catch (e) {
      return Uint8List.fromList(<int>[0xfe, 0xff] + encodeUtf16be(value));
    }
  }

  static Uint8List _date(DateTime date) {
    var utcDate = date.toUtc();
    var year = utcDate.year.toString().padLeft(4, '0');
    var month = utcDate.month.toString().padLeft(2, '0');
    var day = utcDate.day.toString().padLeft(2, '0');
    var hour = utcDate.hour.toString().padLeft(2, '0');
    var minute = utcDate.minute.toString().padLeft(2, '0');
    var second = utcDate.second.toString().padLeft(2, '0');
    return _string('D:$year$month$day$hour$minute${second}Z');
  }

  /// Escape special characters
  /// \ddd Character code ddd (octal)
  void _putTextBytes(PStream s, List<int> b) {
    for (var c in b) {
      switch (c) {
        case 0x0a: // \n Line feed (LF)
          s.putByte(0x5c);
          s.putByte(0x6e);
          break;
        case 0x0d: // \r Carriage return (CR)
          s.putByte(0x5c);
          s.putByte(0x72);
          break;
        case 0x09: // \t Horizontal tab (HT)
          s.putByte(0x5c);
          s.putByte(0x74);
          break;
        case 0x08: // \b Backspace (BS)
          s.putByte(0x5c);
          s.putByte(0x62);
          break;
        case 0x0c: // \f Form feed (FF)
          s.putByte(0x5c);
          s.putByte(0x66);
          break;
        case 0x28: // \( Left parenthesis
          s.putByte(0x5c);
          s.putByte(0x28);
          break;
        case 0x29: // \) Right parenthesis
          s.putByte(0x5c);
          s.putByte(0x29);
          break;
        case 0x5c: // \\ Backslash
          s.putByte(0x5c);
          s.putByte(0x5c);
          break;
        default:
          s.putByte(c);
      }
    }
  }

  /// Returns the ASCII/Unicode code unit corresponding to the hexadecimal digit
  /// [digit].
  int _codeUnitForDigit(int digit) =>digit < 10 ? digit + 0x30 : digit + 0x61 - 10;

  void _output(PStream s, Uint8List value) {

    if(format){
      s.putByte(0x3c);
      for (var byte in value) {
        s.putByte(_codeUnitForDigit((byte & 0xF0) >> 4));
        s.putByte(_codeUnitForDigit(byte & 0x0F));
      }
      s.putByte(0x3e);
    }else{
      s.putByte(40);
      _putTextBytes(s, value);
      s.putByte(41);
    }
  }

  @override
  void output(PStream s) {
    _output(s, value);
  }
}

class PName extends PType {
  const PName(this.value);
  final String value;
  @override
  void output(PStream s) {
    assert(value[0] == '/');
    s.putString(value);
  }
}

class PNull extends PType {
  const PNull();
  @override
  void output(PStream s) {
    s.putString('null');
  }
}

class PIDR extends PType {
  const PIDR(this.ser, this.gen);
  final int ser;
  final int gen;
  @override
  void output(PStream s) {
    s.putString('$ser $gen R');
  }
}

class PArray extends PType {
  PArray([Iterable<PType> values]) {
    if (values != null) {
      this.values.addAll(values);
    }
  }

  factory PArray.fromObjects(List<PObject> objects) {
    return PArray(objects.map<PIDR>((PObject e) => e.idref()).toList());
  }

  factory PArray.fromNum(List<num> list) {
    return PArray(list.map<PNum>((num e) => PNum(e)).toList());
  }

  final List<PType> values = <PType>[];

  void add(PType v) {
    values.add(v);
  }

  @override
  void output(PStream s) {
    s.putString('[');
    if (values.isNotEmpty) {
      for (var n = 0; n < values.length; n++) {
        var val = values[n];
        if (n > 0 &&
            !(val is PName ||
                val is PString ||
                val is PArray ||
                val is PDict)) {
          s.putByte(0x20);
        }
        val.output(s);
      }
    }
    s.putString(']');
  }
}

class PDict extends PType {
  PDict([Map<String, PType> values]) {
    if (values != null) {
      this.values.addAll(values);
    }
  }

  factory PDict.fromObjectMap(Map<String, PObject> objects) {
    return PDict(
      objects.map<String, PIDR>(
        (String key, PObject value) =>
            MapEntry<String, PIDR>(key, value.idref()),
      ),
    );
  }

  final Map<String, PType> values = <String, PType>{};

  bool get isNotEmpty => values.isNotEmpty;

  operator []=(String k, PType v) {
    values[k] = v;
  }

  @override
  void output(PStream s) {
    s.putBytes(const <int>[0x3c, 0x3c]);
    values.forEach((String k, PType v) {
      s.putString(k);
      if (v is PNum || v is PBool || v is PNull || v is PIDR) {
        s.putByte(0x20);
      }
      v.output(s);
    });
    s.putBytes(const <int>[0x3e, 0x3e]);
  }

  bool containsKey(String key) {
    return values.containsKey(key);
  }
}
